import re
from typing import List, Union, Callable, Any

def value_set(values):
    return lambda value: value in values

def matches(pattern: str):
    regex = re.compile(pattern)
    return lambda value: regex.fullmatch(value) is not None

def min_length(min_length: int):
    return lambda value: len(value) >= min_length

def max_length(max_length: int):
    return lambda value: len(value) <= max_length

def multiple_of(multiple_of: Union[int, float]):
    return lambda value: value % multiple_of == 0

def minimum(minimum: Union[int, float], exclusive_minimum: bool):
    if exclusive_minimum:
        return lambda value: value > minimum
    else:
        return lambda value: value >= minimum

def maximum(maximum: Union[int, float], exclusive_maximum: bool):
    if exclusive_maximum:
        return lambda value: value < maximum
    else:
        return lambda value: value <= maximum

def listitem_validator(*validators: Callable[[Any], bool]):
    return lambda value: all([
            all([validator(item) for validator in validators]) 
            for item in value
        ])

def min_items(min_items: int):
    return min_length(min_items)

def max_items(max_items: int):
    return max_length(max_items)

def unique_items():
    return lambda value: len(set(value)) == len(value)



"""GRAPHISOFT
"""
from typing import Dict, Any, List, Tuple, Optional, Union
from urllib.request import Request, urlopen
import json
from archicad.acbasetype import _ACBaseType, _ConstructUnion, _ListBuilder
from archicad.validators import value_set, matches, min_length, max_length, multiple_of, minimum, maximum, listitem_validator, min_items, max_items, unique_items

from .b3000types import NavigatorItemId, LayoutParameters, Subset, FolderParameters, NavigatorItemIdWrapper, ExecutionResult, ElementIdArrayItem, BoundingBox2DOrError, BoundingBox3DOrError, ClassificationSystemId, ClassificationItemArrayItem, ClassificationSystem, PropertyUserId, ClassificationSystemIdArrayItem, ElementClassificationOrError, ClassificationItemIdArrayItem, ClassificationItemOrError, PropertyIdArrayItem, PropertyDefinitionOrError, ClassificationItemId, ElementsOrError, NavigatorTreeId, NavigatorTree, PropertyIdOrError, PropertyValuesOrError, ElementClassification, ElementPropertyValue


class UnsucceededCommandCall(Exception):
    pass


def post_command(req: Request, jsonStr: str) -> Dict[str, Any]:
    response = urlopen(req, jsonStr.encode("UTF-8"))
    result = response.read()
    return json.loads(result)


class Commands:
    """Collection of the ARCHICAD JSON interface commands
    """
    def __init__(self, req: Request):
        assert req is not None
        self.__req = req

    def CloneProjectMapItemToViewMap(self, projectMapNavigatorItemId: NavigatorItemId, parentNavigatorItemId: NavigatorItemId) -> NavigatorItemId:
        """Clones a project map item to the view map.

        Args:
            projectMapNavigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.
            parentNavigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.

        Returns:
            :obj:`NavigatorItemId`: The identifier of a navigator item.

        """
        class CloneProjectMapItemToViewMap_parameters(_ACBaseType):
            __slots__ = ("projectMapNavigatorItemId", "parentNavigatorItemId", )
            def __init__(self, projectMapNavigatorItemId: NavigatorItemId, parentNavigatorItemId: NavigatorItemId):
                self.projectMapNavigatorItemId: NavigatorItemId = projectMapNavigatorItemId
                self.parentNavigatorItemId: NavigatorItemId = parentNavigatorItemId

        CloneProjectMapItemToViewMap_parameters.get_classinfo().add_field('projectMapNavigatorItemId', NavigatorItemId)
        CloneProjectMapItemToViewMap_parameters.get_classinfo().add_field('parentNavigatorItemId', NavigatorItemId)

        result = post_command(self.__req, json.dumps({"command": "API.CloneProjectMapItemToViewMap", "parameters": CloneProjectMapItemToViewMap_parameters(projectMapNavigatorItemId, parentNavigatorItemId).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        return NavigatorItemId(**result["result"]["createdNavigatorItemId"])

    def CreateLayout(self, layoutName: str, layoutParameters: LayoutParameters, masterNavigatorItemId: NavigatorItemId, parentNavigatorItemId: NavigatorItemId) -> NavigatorItemId:
        """Creates a new layout.

        Args:
            layoutName (:obj:`str`): The name of the layout.
            layoutParameters (:obj:`LayoutParameters`): The parameters of the layout.
            masterNavigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.
            parentNavigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.

        Returns:
            :obj:`NavigatorItemId`: The identifier of a navigator item.

        """
        class CreateLayout_parameters(_ACBaseType):
            __slots__ = ("layoutName", "layoutParameters", "masterNavigatorItemId", "parentNavigatorItemId", )
            def __init__(self, layoutName: str, layoutParameters: LayoutParameters, masterNavigatorItemId: NavigatorItemId, parentNavigatorItemId: NavigatorItemId):
                self.layoutName: str = layoutName
                self.layoutParameters: LayoutParameters = layoutParameters
                self.masterNavigatorItemId: NavigatorItemId = masterNavigatorItemId
                self.parentNavigatorItemId: NavigatorItemId = parentNavigatorItemId

        CreateLayout_parameters.get_classinfo().add_field('layoutName', str, min_length(1))
        CreateLayout_parameters.get_classinfo().add_field('layoutParameters', LayoutParameters)
        CreateLayout_parameters.get_classinfo().add_field('masterNavigatorItemId', NavigatorItemId)
        CreateLayout_parameters.get_classinfo().add_field('parentNavigatorItemId', NavigatorItemId)

        result = post_command(self.__req, json.dumps({"command": "API.CreateLayout", "parameters": CreateLayout_parameters(layoutName, layoutParameters, masterNavigatorItemId, parentNavigatorItemId).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        return NavigatorItemId(**result["result"]["createdNavigatorItemId"])

    def CreateLayoutSubset(self, subsetParameters: Subset, parentNavigatorItemId: NavigatorItemId) -> NavigatorItemId:
        """Creates a new layout subset.

        Args:
            subsetParameters (:obj:`Subset`): Provides a way to assign IDs to the layouts contained in the subset.
            parentNavigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.

        Returns:
            :obj:`NavigatorItemId`: The identifier of a navigator item.

        """
        class CreateLayoutSubset_parameters(_ACBaseType):
            __slots__ = ("subsetParameters", "parentNavigatorItemId", )
            def __init__(self, subsetParameters: Subset, parentNavigatorItemId: NavigatorItemId):
                self.subsetParameters: Subset = subsetParameters
                self.parentNavigatorItemId: NavigatorItemId = parentNavigatorItemId

        CreateLayoutSubset_parameters.get_classinfo().add_field('subsetParameters', Subset)
        CreateLayoutSubset_parameters.get_classinfo().add_field('parentNavigatorItemId', NavigatorItemId)

        result = post_command(self.__req, json.dumps({"command": "API.CreateLayoutSubset", "parameters": CreateLayoutSubset_parameters(subsetParameters, parentNavigatorItemId).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        return NavigatorItemId(**result["result"]["createdSubsetId"])

    def CreateViewMapFolder(self, folderParameters: FolderParameters, parentNavigatorItemId: Optional[NavigatorItemId] = None, previousNavigatorItemId: Optional[NavigatorItemId] = None) -> NavigatorItemId:
        """Creates a view folder item at the given position in the navigator tree.

        Args:
            folderParameters (:obj:`FolderParameters`): The parameters of a folder.
            parentNavigatorItemId (:obj:`NavigatorItemId`, optional): The identifier of a navigator item.
            previousNavigatorItemId (:obj:`NavigatorItemId`, optional): The identifier of a navigator item.

        Returns:
            :obj:`NavigatorItemId`: The identifier of a navigator item.

        """
        class CreateViewMapFolder_parameters(_ACBaseType):
            __slots__ = ("folderParameters", "parentNavigatorItemId", "previousNavigatorItemId", )
            def __init__(self, folderParameters: FolderParameters, parentNavigatorItemId: Optional[NavigatorItemId] = None, previousNavigatorItemId: Optional[NavigatorItemId] = None):
                self.folderParameters: FolderParameters = folderParameters
                self.parentNavigatorItemId: Optional[NavigatorItemId] = parentNavigatorItemId
                self.previousNavigatorItemId: Optional[NavigatorItemId] = previousNavigatorItemId

        CreateViewMapFolder_parameters.get_classinfo().add_field('folderParameters', FolderParameters)
        CreateViewMapFolder_parameters.get_classinfo().add_field('parentNavigatorItemId', Optional[NavigatorItemId])
        CreateViewMapFolder_parameters.get_classinfo().add_field('previousNavigatorItemId', Optional[NavigatorItemId])

        result = post_command(self.__req, json.dumps({"command": "API.CreateViewMapFolder", "parameters": CreateViewMapFolder_parameters(folderParameters, parentNavigatorItemId, previousNavigatorItemId).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        return NavigatorItemId(**result["result"]["createdFolderNavigatorItemId"])

    def DeleteNavigatorItems(self, navigatorItemIds: List[NavigatorItemIdWrapper]) -> List[ExecutionResult]:
        """Deletes items from navigator tree.

        Args:
            navigatorItemIds (:obj:`list` of :obj:`NavigatorItemIdWrapper`): List of navigator item identifiers.

        Returns:
            :obj:`list` of :obj:`ExecutionResult`: Execution result for each input case.

        """
        class DeleteNavigatorItems_parameters(_ACBaseType):
            __slots__ = ("navigatorItemIds", )
            def __init__(self, navigatorItemIds: List[NavigatorItemIdWrapper]):
                self.navigatorItemIds: List[NavigatorItemIdWrapper] = navigatorItemIds

        DeleteNavigatorItems_parameters.get_classinfo().add_field('navigatorItemIds', List[NavigatorItemIdWrapper])

        result = post_command(self.__req, json.dumps({"command": "API.DeleteNavigatorItems", "parameters": DeleteNavigatorItems_parameters(navigatorItemIds).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        executionResultsListBuilder = _ListBuilder(ExecutionResult)
        return executionResultsListBuilder(result["result"]["executionResults"])

    def Get2DBoundingBoxes(self, elements: List[ElementIdArrayItem]) -> List[BoundingBox2DOrError]:
        """Get the 2D bounding box of elements identified by their GUIDs. The bounding box is calculated from the global origin on the floor plan view. The output is the array of the bounding boxes respective to the input GUIDs. Only works for elements detailed in <i>Element Information</i>.

        Args:
            elements (:obj:`list` of :obj:`ElementIdArrayItem`): List of the elements.

        Returns:
            :obj:`list` of :obj:`BoundingBox2DOrError`: List of element's 2D bounding boxes.

        """
        class Get2DBoundingBoxes_parameters(_ACBaseType):
            __slots__ = ("elements", )
            def __init__(self, elements: List[ElementIdArrayItem]):
                self.elements: List[ElementIdArrayItem] = elements

        Get2DBoundingBoxes_parameters.get_classinfo().add_field('elements', List[ElementIdArrayItem])

        result = post_command(self.__req, json.dumps({"command": "API.Get2DBoundingBoxes", "parameters": Get2DBoundingBoxes_parameters(elements).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        boundingBoxes2DListBuilder = _ListBuilder(BoundingBox2DOrError)
        return boundingBoxes2DListBuilder(result["result"]["boundingBoxes2D"])

    def Get3DBoundingBoxes(self, elements: List[ElementIdArrayItem]) -> List[BoundingBox3DOrError]:
        """Get the 3D bounding box of elements identified by their GUIDs. The bounding box is calculated from the global origin in the 3D view. The output is the array of the bounding boxes respective to the input GUIDs. Only works for elements detailed in <i>Element Information</i>.

        Args:
            elements (:obj:`list` of :obj:`ElementIdArrayItem`): List of the elements.

        Returns:
            :obj:`list` of :obj:`BoundingBox3DOrError`: List of element's 3D bounding boxes

        """
        class Get3DBoundingBoxes_parameters(_ACBaseType):
            __slots__ = ("elements", )
            def __init__(self, elements: List[ElementIdArrayItem]):
                self.elements: List[ElementIdArrayItem] = elements

        Get3DBoundingBoxes_parameters.get_classinfo().add_field('elements', List[ElementIdArrayItem])

        result = post_command(self.__req, json.dumps({"command": "API.Get3DBoundingBoxes", "parameters": Get3DBoundingBoxes_parameters(elements).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        boundingBoxes3DListBuilder = _ListBuilder(BoundingBox3DOrError)
        return boundingBoxes3DListBuilder(result["result"]["boundingBoxes3D"])

    def GetAllClassificationsInSystem(self, classificationSystemId: ClassificationSystemId) -> List[ClassificationItemArrayItem]:
        """Returns the tree of classifications in the given classification system.

        Args:
            classificationSystemId (:obj:`ClassificationSystemId`): The identifier of a classification system.

        Returns:
            :obj:`list` of :obj:`ClassificationItemArrayItem`: The list of classification items.

        """
        class GetAllClassificationsInSystem_parameters(_ACBaseType):
            __slots__ = ("classificationSystemId", )
            def __init__(self, classificationSystemId: ClassificationSystemId):
                self.classificationSystemId: ClassificationSystemId = classificationSystemId

        GetAllClassificationsInSystem_parameters.get_classinfo().add_field('classificationSystemId', ClassificationSystemId)

        result = post_command(self.__req, json.dumps({"command": "API.GetAllClassificationsInSystem", "parameters": GetAllClassificationsInSystem_parameters(classificationSystemId).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        classificationItemsListBuilder = _ListBuilder(ClassificationItemArrayItem)
        return classificationItemsListBuilder(result["result"]["classificationItems"])

    def GetAllClassificationSystems(self) -> List[ClassificationSystem]:
        """Returns the list of available classification systems.

        Returns:
            :obj:`list` of :obj:`ClassificationSystem`: The list of classification systems.

        """

        result = post_command(self.__req, json.dumps({"command": "API.GetAllClassificationSystems"}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        classificationSystemsListBuilder = _ListBuilder(ClassificationSystem)
        return classificationSystemsListBuilder(result["result"]["classificationSystems"])

    def GetAllElements(self) -> List[ElementIdArrayItem]:
        """Returns the identifier of every element in the current plan.

        Returns:
            :obj:`list` of :obj:`ElementIdArrayItem`: List of the elements.

        """

        result = post_command(self.__req, json.dumps({"command": "API.GetAllElements"}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        elementsListBuilder = _ListBuilder(ElementIdArrayItem)
        return elementsListBuilder(result["result"]["elements"])

    def GetAllPropertyNames(self) -> List[PropertyUserId]:
        """Returns the human-readable names of available Property definitions for debug and development purposes.

        Returns:
            :obj:`list` of :obj:`PropertyUserId`: List of PropertyUserId objects.

        """

        result = post_command(self.__req, json.dumps({"command": "API.GetAllPropertyNames"}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        propertiesListBuilder = _ListBuilder(PropertyUserId)
        return propertiesListBuilder(result["result"]["properties"])

    def GetClassificationsOfElements(self, elements: List[ElementIdArrayItem], classificationSystemIds: List[ClassificationSystemIdArrayItem]) -> List[ElementClassificationOrError]:
        """Returns the classification of the given elements in the given classification systems.

        Args:
            elements (:obj:`list` of :obj:`ElementIdArrayItem`): List of the elements.
            classificationSystemIds (:obj:`list` of :obj:`ClassificationSystemIdArrayItem`): The list of classification system identifiers.

        Returns:
            :obj:`list` of :obj:`ElementClassificationOrError`: The list of element classification identifiers or errors.

        """
        class GetClassificationsOfElements_parameters(_ACBaseType):
            __slots__ = ("elements", "classificationSystemIds", )
            def __init__(self, elements: List[ElementIdArrayItem], classificationSystemIds: List[ClassificationSystemIdArrayItem]):
                self.elements: List[ElementIdArrayItem] = elements
                self.classificationSystemIds: List[ClassificationSystemIdArrayItem] = classificationSystemIds

        GetClassificationsOfElements_parameters.get_classinfo().add_field('elements', List[ElementIdArrayItem])
        GetClassificationsOfElements_parameters.get_classinfo().add_field('classificationSystemIds', List[ClassificationSystemIdArrayItem])

        result = post_command(self.__req, json.dumps({"command": "API.GetClassificationsOfElements", "parameters": GetClassificationsOfElements_parameters(elements, classificationSystemIds).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        elementClassificationsListBuilder = _ListBuilder(ElementClassificationOrError)
        return elementClassificationsListBuilder(result["result"]["elementClassifications"])

    def GetDetailsOfClassificationItems(self, classificationItemIds: List[ClassificationItemIdArrayItem]) -> List[ClassificationItemOrError]:
        """Returns the details of classification items.

        Args:
            classificationItemIds (:obj:`list` of :obj:`ClassificationItemIdArrayItem`): The list of classification item identifiers.

        Returns:
            :obj:`list` of :obj:`ClassificationItemOrError`: The list of classification items or errors.

        """
        class GetDetailsOfClassificationItems_parameters(_ACBaseType):
            __slots__ = ("classificationItemIds", )
            def __init__(self, classificationItemIds: List[ClassificationItemIdArrayItem]):
                self.classificationItemIds: List[ClassificationItemIdArrayItem] = classificationItemIds

        GetDetailsOfClassificationItems_parameters.get_classinfo().add_field('classificationItemIds', List[ClassificationItemIdArrayItem])

        result = post_command(self.__req, json.dumps({"command": "API.GetDetailsOfClassificationItems", "parameters": GetDetailsOfClassificationItems_parameters(classificationItemIds).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        classificationItemsListBuilder = _ListBuilder(ClassificationItemOrError)
        return classificationItemsListBuilder(result["result"]["classificationItems"])

    def GetDetailsOfProperties(self, properties: List[PropertyIdArrayItem]) -> List[PropertyDefinitionOrError]:
        """Returns the details of property definitions.

        Args:
            properties (:obj:`list` of :obj:`PropertyIdArrayItem`): The list of property identifiers.

        Returns:
            :obj:`list` of :obj:`PropertyDefinitionOrError`: The list of property definitions or errors.

        """
        class GetDetailsOfProperties_parameters(_ACBaseType):
            __slots__ = ("properties", )
            def __init__(self, properties: List[PropertyIdArrayItem]):
                self.properties: List[PropertyIdArrayItem] = properties

        GetDetailsOfProperties_parameters.get_classinfo().add_field('properties', List[PropertyIdArrayItem])

        result = post_command(self.__req, json.dumps({"command": "API.GetDetailsOfProperties", "parameters": GetDetailsOfProperties_parameters(properties).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        propertyDefinitionsListBuilder = _ListBuilder(PropertyDefinitionOrError)
        return propertyDefinitionsListBuilder(result["result"]["propertyDefinitions"])

    def GetElementsByClassification(self, classificationItemId: ClassificationItemId) -> List[ElementIdArrayItem]:
        """Returns the identifier of every element with the given classification identifier.

        Args:
            classificationItemId (:obj:`ClassificationItemId`): The identifier of a classification item.

        Returns:
            :obj:`list` of :obj:`ElementIdArrayItem`: List of the elements.

        """
        class GetElementsByClassification_parameters(_ACBaseType):
            __slots__ = ("classificationItemId", )
            def __init__(self, classificationItemId: ClassificationItemId):
                self.classificationItemId: ClassificationItemId = classificationItemId

        GetElementsByClassification_parameters.get_classinfo().add_field('classificationItemId', ClassificationItemId)

        result = post_command(self.__req, json.dumps({"command": "API.GetElementsByClassification", "parameters": GetElementsByClassification_parameters(classificationItemId).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        elementsListBuilder = _ListBuilder(ElementIdArrayItem)
        return elementsListBuilder(result["result"]["elements"])

    def GetElementsByType(self, elementType: str) -> List[ElementIdArrayItem]:
        """Returns the identifier of every element of the given type on the plan.

        Args:
            elementType (:obj:`str`): The type of an element.

        Returns:
            :obj:`list` of :obj:`ElementIdArrayItem`: List of the elements.

        """
        class GetElementsByType_parameters(_ACBaseType):
            __slots__ = ("elementType", )
            def __init__(self, elementType: str):
                self.elementType: str = elementType

        GetElementsByType_parameters.get_classinfo().add_field('elementType', str, value_set(['Wall', 'Column', 'Beam', 'Window', 'Door', 'Object', 'Lamp', 'Slab', 'Roof', 'Mesh', 'Zone', 'CurtainWall', 'Shell', 'Skylight', 'Morph', 'Stair', 'Railing', 'Opening']))

        result = post_command(self.__req, json.dumps({"command": "API.GetElementsByType", "parameters": GetElementsByType_parameters(elementType).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        elementsListBuilder = _ListBuilder(ElementIdArrayItem)
        return elementsListBuilder(result["result"]["elements"])

    def GetElementsRelatedToZones(self, zones: List[ElementIdArrayItem], elementTypes: Optional[List[str]] = None) -> List[ElementsOrError]:
        """Returns related elements of the given zones. The related elements will be grouped by type. If multiple zones was given, then the order of the returned list is that of the given zones.

        Args:
            zones (:obj:`list` of :obj:`ElementIdArrayItem`): List of the elements.
            elementTypes (:obj:`list` of :obj:`str`, optional): List of element types.

        Returns:
            :obj:`list` of :obj:`ElementsOrError`: List of elements or error.

        """
        class GetElementsRelatedToZones_parameters(_ACBaseType):
            __slots__ = ("zones", "elementTypes", )
            def __init__(self, zones: List[ElementIdArrayItem], elementTypes: Optional[List[str]] = None):
                self.zones: List[ElementIdArrayItem] = zones
                self.elementTypes: Optional[List[str]] = elementTypes

        GetElementsRelatedToZones_parameters.get_classinfo().add_field('zones', List[ElementIdArrayItem])
        GetElementsRelatedToZones_parameters.get_classinfo().add_field('elementTypes', Optional[List[str]], listitem_validator(value_set(['Wall', 'Column', 'Beam', 'Window', 'Door', 'Object', 'Lamp', 'Slab', 'Roof', 'Mesh', 'Zone', 'CurtainWall', 'Shell', 'Skylight', 'Morph', 'Stair', 'Railing', 'Opening'])))

        result = post_command(self.__req, json.dumps({"command": "API.GetElementsRelatedToZones", "parameters": GetElementsRelatedToZones_parameters(zones, elementTypes).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        elementsRelatedToZonesListBuilder = _ListBuilder(ElementsOrError)
        return elementsRelatedToZonesListBuilder(result["result"]["elementsRelatedToZones"])

    def GetLayoutSettings(self, layoutNavigatorItemId: NavigatorItemId) -> LayoutParameters:
        """Returns the parameters (settings) of the given layout.

        Args:
            layoutNavigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.

        Returns:
            :obj:`LayoutParameters`: The parameters of the layout.

        """
        class GetLayoutSettings_parameters(_ACBaseType):
            __slots__ = ("layoutNavigatorItemId", )
            def __init__(self, layoutNavigatorItemId: NavigatorItemId):
                self.layoutNavigatorItemId: NavigatorItemId = layoutNavigatorItemId

        GetLayoutSettings_parameters.get_classinfo().add_field('layoutNavigatorItemId', NavigatorItemId)

        result = post_command(self.__req, json.dumps({"command": "API.GetLayoutSettings", "parameters": GetLayoutSettings_parameters(layoutNavigatorItemId).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        return LayoutParameters(**result["result"]["layoutParameters"])

    def GetNavigatorItemTree(self, navigatorTreeId: NavigatorTreeId) -> NavigatorTree:
        """Returns the tree of navigator items.

        Args:
            navigatorTreeId (:obj:`NavigatorTreeId`): The identifier of a navigator item tree.

        Returns:
            :obj:`NavigatorTree`: A tree of navigator items.

        """
        class GetNavigatorItemTree_parameters(_ACBaseType):
            __slots__ = ("navigatorTreeId", )
            def __init__(self, navigatorTreeId: NavigatorTreeId):
                self.navigatorTreeId: NavigatorTreeId = navigatorTreeId

        GetNavigatorItemTree_parameters.get_classinfo().add_field('navigatorTreeId', NavigatorTreeId)

        result = post_command(self.__req, json.dumps({"command": "API.GetNavigatorItemTree", "parameters": GetNavigatorItemTree_parameters(navigatorTreeId).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        return NavigatorTree(**result["result"]["navigatorTree"])

    def GetProductInfo(self) -> Tuple[int, int, str]:
        """Accesses the version information from the running ARCHICAD.

        Returns:
            :obj:`int`: The version of the running ARCHICAD.
            :obj:`int`: The build number of the running ARCHICAD.
            :obj:`str`: The language code of the running ARCHICAD.

        """

        result = post_command(self.__req, json.dumps({"command": "API.GetProductInfo"}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        return result["result"]["version"], result["result"]["buildNumber"], result["result"]["languageCode"]

    def GetPropertyIds(self, properties: List[PropertyUserId]) -> List[PropertyIdOrError]:
        """Returns the identifiers of property definitions for the requested property names.

        Args:
            properties (:obj:`list` of :obj:`PropertyUserId`): List of PropertyUserId objects.

        Returns:
            :obj:`list` of :obj:`PropertyIdOrError`: The list of property identifiers or errors.

        """
        class GetPropertyIds_parameters(_ACBaseType):
            __slots__ = ("properties", )
            def __init__(self, properties: List[PropertyUserId]):
                self.properties: List[PropertyUserId] = properties

        GetPropertyIds_parameters.get_classinfo().add_field('properties', List[PropertyUserId])

        result = post_command(self.__req, json.dumps({"command": "API.GetPropertyIds", "parameters": GetPropertyIds_parameters(properties).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        propertiesListBuilder = _ListBuilder(PropertyIdOrError)
        return propertiesListBuilder(result["result"]["properties"])

    def GetPropertyValuesOfElements(self, elements: List[ElementIdArrayItem], properties: List[PropertyIdArrayItem]) -> List[PropertyValuesOrError]:
        """Returns the property values of the elements for the given property.

        Args:
            elements (:obj:`list` of :obj:`ElementIdArrayItem`): List of the elements.
            properties (:obj:`list` of :obj:`PropertyIdArrayItem`): The list of property identifiers.

        Returns:
            :obj:`list` of :obj:`PropertyValuesOrError`: List of property value lists.

        """
        class GetPropertyValuesOfElements_parameters(_ACBaseType):
            __slots__ = ("elements", "properties", )
            def __init__(self, elements: List[ElementIdArrayItem], properties: List[PropertyIdArrayItem]):
                self.elements: List[ElementIdArrayItem] = elements
                self.properties: List[PropertyIdArrayItem] = properties

        GetPropertyValuesOfElements_parameters.get_classinfo().add_field('elements', List[ElementIdArrayItem])
        GetPropertyValuesOfElements_parameters.get_classinfo().add_field('properties', List[PropertyIdArrayItem])

        result = post_command(self.__req, json.dumps({"command": "API.GetPropertyValuesOfElements", "parameters": GetPropertyValuesOfElements_parameters(elements, properties).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        propertyValuesForElementsListBuilder = _ListBuilder(PropertyValuesOrError)
        return propertyValuesForElementsListBuilder(result["result"]["propertyValuesForElements"])

    def GetPublisherSetNames(self) -> List[str]:
        """Returns the names of available publisher sets.

        Returns:
            :obj:`list` of :obj:`str`: The names of available publisher sets.

        """

        result = post_command(self.__req, json.dumps({"command": "API.GetPublisherSetNames"}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        publisherSetNamesListBuilder = _ListBuilder(str)
        return publisherSetNamesListBuilder(result["result"]["publisherSetNames"])

    def IsAlive(self) -> bool:
        """Checks if the ARCHICAD connection is alive.

        Returns:
            :obj:`bool`: Returns true if the connection is alive.

        """

        result = post_command(self.__req, json.dumps({"command": "API.IsAlive"}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        return result["result"]["isAlive"]

    def MoveNavigatorItem(self, navigatorItemIdToMove: NavigatorItemId, parentNavigatorItemId: NavigatorItemId, previousNavigatorItemId: Optional[NavigatorItemId] = None):
        """Moves the given navigator item under the <i>parentNavigatorItemId</i> in the navigator tree. If <i>previousNavigatorItemId</i> is not given then inserts it at the first place under the new parent. If it is given then inserts it after this navigator item.

        Args:
            navigatorItemIdToMove (:obj:`NavigatorItemId`): The identifier of a navigator item.
            parentNavigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.
            previousNavigatorItemId (:obj:`NavigatorItemId`, optional): The identifier of a navigator item.

        """
        class MoveNavigatorItem_parameters(_ACBaseType):
            __slots__ = ("navigatorItemIdToMove", "parentNavigatorItemId", "previousNavigatorItemId", )
            def __init__(self, navigatorItemIdToMove: NavigatorItemId, parentNavigatorItemId: NavigatorItemId, previousNavigatorItemId: Optional[NavigatorItemId] = None):
                self.navigatorItemIdToMove: NavigatorItemId = navigatorItemIdToMove
                self.parentNavigatorItemId: NavigatorItemId = parentNavigatorItemId
                self.previousNavigatorItemId: Optional[NavigatorItemId] = previousNavigatorItemId

        MoveNavigatorItem_parameters.get_classinfo().add_field('navigatorItemIdToMove', NavigatorItemId)
        MoveNavigatorItem_parameters.get_classinfo().add_field('parentNavigatorItemId', NavigatorItemId)
        MoveNavigatorItem_parameters.get_classinfo().add_field('previousNavigatorItemId', Optional[NavigatorItemId])

        result = post_command(self.__req, json.dumps({"command": "API.MoveNavigatorItem", "parameters": MoveNavigatorItem_parameters(navigatorItemIdToMove, parentNavigatorItemId, previousNavigatorItemId).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)


    def RenameNavigatorItem(self, navigatorItemId: NavigatorItemId, newName: Optional[str] = None, newId: Optional[str] = None):
        """Renames an existing navigator item by specifying either the name or the ID, or both.

        Args:
            navigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.
            newName (:obj:`str`, optional): New name of the navigator item.
            newId (:obj:`str`, optional): New ID of the navigator item.

        """
        class RenameNavigatorItem_parameters1(_ACBaseType):
            __slots__ = ("navigatorItemId", "newName", )
            def __init__(self, navigatorItemId: NavigatorItemId, newName: str):
                self.navigatorItemId: NavigatorItemId = navigatorItemId
                self.newName: str = newName

        RenameNavigatorItem_parameters1.get_classinfo().add_field('navigatorItemId', NavigatorItemId)
        RenameNavigatorItem_parameters1.get_classinfo().add_field('newName', str)

        class RenameNavigatorItem_parameters2(_ACBaseType):
            __slots__ = ("navigatorItemId", "newId", )
            def __init__(self, navigatorItemId: NavigatorItemId, newId: str):
                self.navigatorItemId: NavigatorItemId = navigatorItemId
                self.newId: str = newId

        RenameNavigatorItem_parameters2.get_classinfo().add_field('navigatorItemId', NavigatorItemId)
        RenameNavigatorItem_parameters2.get_classinfo().add_field('newId', str)

        class RenameNavigatorItem_parameters3(_ACBaseType):
            __slots__ = ("navigatorItemId", "newName", "newId", )
            def __init__(self, navigatorItemId: NavigatorItemId, newName: str, newId: str):
                self.navigatorItemId: NavigatorItemId = navigatorItemId
                self.newName: str = newName
                self.newId: str = newId

        RenameNavigatorItem_parameters3.get_classinfo().add_field('navigatorItemId', NavigatorItemId)
        RenameNavigatorItem_parameters3.get_classinfo().add_field('newName', str)
        RenameNavigatorItem_parameters3.get_classinfo().add_field('newId', str)

        parametersConstructUnion = _ConstructUnion(Union[RenameNavigatorItem_parameters1, RenameNavigatorItem_parameters2, RenameNavigatorItem_parameters3])
        paramatersObject = parametersConstructUnion(navigatorItemId=navigatorItemId, newName=newName, newId=newId)

        result = post_command(self.__req, json.dumps({"command": "API.RenameNavigatorItem", "parameters": paramatersObject.to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)


    def SetClassificationsOfElements(self, elementClassifications: List[ElementClassification]) -> List[ExecutionResult]:
        """Sets the classifications of elements. In order to set the classification of an element to unclassified, omit the classificationItemId field.

        Args:
            elementClassifications (:obj:`list` of :obj:`ElementClassification`): The list of element classification identifiers.

        Returns:
            :obj:`list` of :obj:`ExecutionResult`: Execution result for each input case.

        """
        class SetClassificationsOfElements_parameters(_ACBaseType):
            __slots__ = ("elementClassifications", )
            def __init__(self, elementClassifications: List[ElementClassification]):
                self.elementClassifications: List[ElementClassification] = elementClassifications

        SetClassificationsOfElements_parameters.get_classinfo().add_field('elementClassifications', List[ElementClassification])

        result = post_command(self.__req, json.dumps({"command": "API.SetClassificationsOfElements", "parameters": SetClassificationsOfElements_parameters(elementClassifications).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        executionResultsListBuilder = _ListBuilder(ExecutionResult)
        return executionResultsListBuilder(result["result"]["executionResults"])

    def SetLayoutSettings(self, layoutParameters: LayoutParameters, layoutNavigatorItemId: NavigatorItemId):
        """Sets the parameters (settings) of the given layout.

        Args:
            layoutParameters (:obj:`LayoutParameters`): The parameters of the layout.
            layoutNavigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.

        """
        class SetLayoutSettings_parameters(_ACBaseType):
            __slots__ = ("layoutParameters", "layoutNavigatorItemId", )
            def __init__(self, layoutParameters: LayoutParameters, layoutNavigatorItemId: NavigatorItemId):
                self.layoutParameters: LayoutParameters = layoutParameters
                self.layoutNavigatorItemId: NavigatorItemId = layoutNavigatorItemId

        SetLayoutSettings_parameters.get_classinfo().add_field('layoutParameters', LayoutParameters)
        SetLayoutSettings_parameters.get_classinfo().add_field('layoutNavigatorItemId', NavigatorItemId)

        result = post_command(self.__req, json.dumps({"command": "API.SetLayoutSettings", "parameters": SetLayoutSettings_parameters(layoutParameters, layoutNavigatorItemId).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)


    def SetPropertyValuesOfElements(self, elementPropertyValues: List[ElementPropertyValue]) -> List[ExecutionResult]:
        """Sets the property values of elements.

        Args:
            elementPropertyValues (:obj:`list` of :obj:`ElementPropertyValue`): List of element property values.

        Returns:
            :obj:`list` of :obj:`ExecutionResult`: Execution result for each input case.

        """
        class SetPropertyValuesOfElements_parameters(_ACBaseType):
            __slots__ = ("elementPropertyValues", )
            def __init__(self, elementPropertyValues: List[ElementPropertyValue]):
                self.elementPropertyValues: List[ElementPropertyValue] = elementPropertyValues

        SetPropertyValuesOfElements_parameters.get_classinfo().add_field('elementPropertyValues', List[ElementPropertyValue])

        result = post_command(self.__req, json.dumps({"command": "API.SetPropertyValuesOfElements", "parameters": SetPropertyValuesOfElements_parameters(elementPropertyValues).to_dict()}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        executionResultsListBuilder = _ListBuilder(ExecutionResult)
        return executionResultsListBuilder(result["result"]["executionResults"])




import sys
import importlib
import subprocess


def _check_module(module: str, install_missing_module: bool = True):
    try:
        importlib.import_module(module)
    except ImportError:
        if install_missing_module:
            subprocess.check_call([sys.executable, "-m", "pip", "install", module])
        else:
            print(f"Missing module: '{module}'")

def handle_dependencies(*module_names: str, install_missing_modules: bool = True):
    """ Check the module names' importability and install them if needed.

    Args:
        *module_names: Variable length module name list.
        install_missing_modules (:obj:`bool`): If True then it installs the missing modules else print that the module is missing. Defaults to True.

    """
    for module_name in module_names:
        _check_module(module_name, install_missing_modules)




import os, sys, subprocess
from typing import Optional, Union, Tuple, List, Callable
from archicad.releases.ac24.b3000types import *
from archicad.releases.ac24.b3000commands import *


def _find_in_tree(treeRootItem, itemattr, childrenattr, criterion) -> list:
    result = []
    if criterion(treeRootItem):
        result.append(treeRootItem)
    if treeRootItem.__getattribute__(childrenattr):
        for node in treeRootItem.__getattribute__(childrenattr):
            result.extend(_find_in_tree(node.__getattribute__(itemattr), itemattr, childrenattr, criterion))
    return result


class Utilities:
    """ Utility functions for the archicad module.
    """
    def __init__(self, actypes: Types, accommands: Commands):
        self.actypes = actypes
        self.accommands = accommands

    @staticmethod
    def OpenFile(filepath: str):
        """Opens the given file with the default application."""
        if sys.platform == "win32":
            os.startfile(filepath)
        else:
            subprocess.call(["open" if sys.platform == "darwin" else "xdg-open", filepath])

    @staticmethod
    def FindInNavigatorItemTree(treeRootItem: NavigatorItem,
                                criterion: Callable[[NavigatorItem], bool]) -> List[NavigatorItem]:
        """Finds items in a navigator tree.
        
        Args:
            treeRootItem (:obj:`NavigatorItem`): The root item of the navigator tree.
            criterion (Callable[[NavigatorItem], bool]): The criterion function.
        
        Returns:
            :obj:`List[NavigatorItem]`: The list of navigator items, which fulfill the criterion function.
        """
        return _find_in_tree(treeRootItem, 'navigatorItem', 'children', criterion)

    @staticmethod
    def FindInClassificationItemTree(treeRootItem: ClassificationItemInTree,
                                     criterion: Callable[[ClassificationItemInTree], bool]) -> List[ClassificationItemInTree]:
        """Finds items in a navigator tree.
        
        Args:
            treeRootItem (:obj:`ClassificationItemInTree`): The root item of the classification tree.
            criterion (Callable[[ClassificationItemInTree], bool]): The criterion function.
        
        Returns:
            :obj:`List[ClassificationItemInTree]`: The list of classification items, which fulfill the criterion function.
        """
        return _find_in_tree(treeRootItem, 'classificationItem', 'children', criterion)

    def FindClassificationItemInSystem(self, system_name: str, item_id: str) -> Optional[ClassificationItemId]:
        """Finds the classification item in a system."""
        system_id = next(system.classificationSystemId for system in
                            self.accommands.GetAllClassificationSystems() if system.name == system_name)
        classifications_tree = self.accommands.GetAllClassificationsInSystem(system_id)
        for tree in classifications_tree:
            classification_ids = Utilities.FindInClassificationItemTree(tree.classificationItem, lambda c: c.id == item_id)
            assert len(classification_ids) <= 1
            if classification_ids:
                return classification_ids[0]
        return None
    
    def GetBuiltInPropertyId(self, name: str) -> PropertyId:
        """Returns the PropertyId of the corresponding built-in property."""
        return self.accommands.GetPropertyIds([BuiltInPropertyUserId(name)])[0].propertyId

    def GetUserDefinedPropertyId(self, groupName: str, name: str) -> PropertyId:
        """Returns the PropertyId of the corresponding user-defined property."""
        return self.accommands.GetPropertyIds([UserDefinedPropertyUserId([groupName, name])])[0].propertyId



import unittest
from unittest.mock import patch, ANY

from archicad.versioning import _Versioning, os, importlib


class TestVersioning(unittest.TestCase):
    @patch('archicad.versioning._Versioning.load')
    def test_init(self, mocked_load):
        self.assertRaises(AssertionError, _Versioning, -24, 2310, ANY)
        self.assertRaises(AssertionError, _Versioning, 0, 2310, ANY)
        self.assertRaises(AssertionError, _Versioning, 23, 2310, ANY)
        self.assertIsInstance(_Versioning(24, 2310, ANY), _Versioning)
        self.assertIsInstance(_Versioning(50, 2310, ANY), _Versioning)

    @patch('archicad.versioning.importlib.import_module')
    @patch('archicad.versioning._Versioning.discover')
    def test_load(self, mocked_discover, mocked_import_module):
        self.assertIs(mocked_discover, _Versioning.discover)
        self.assertIs(mocked_import_module, importlib.import_module)
        
        def types():
            return 'Imported Types' 
        def commands(request):
            return 'Imported Commands' 
        def utils(types, commands):
            return 'Imported Utilities' 
        import_module_side_effect = [
            type('obj', (object,), {'Types': types}),
            type('obj', (object,), {'Commands': commands}),
            type('obj', (object,), {'Utilities': utils})
        ]
        mocked_discover.return_value = [25, 2310]
        mocked_import_module.side_effect = import_module_side_effect

        test_version = _Versioning(25, 2320,ANY)

        self.assertEqual(test_version.types, 'Imported Types')
        self.assertEqual(test_version.commands, 'Imported Commands')
        self.assertEqual(test_version.utilities, 'Imported Utilities')

    @patch('archicad.versioning.importlib.import_module')
    @patch('archicad.versioning.os.scandir')
    def test_discover(self, mocked_os_scandir, mocked_import_module):
        self.assertIs(mocked_import_module, importlib.import_module)
        self.assertIs(mocked_os_scandir, os.scandir)

        def returnTrue():
            return True
        mocked_os_scandir.return_value = None
        mocked_import_module.return_value = type('obj', (object,), {'__path__':['random_path_string',]})
        
        os_scandir_side_effect_values = [[
                type('obj', (object,), {'name':'ac25', 'is_dir': returnTrue}),
                type('obj', (object,), {'name':'ac24', 'is_dir': returnTrue}),
                type('obj', (object,), {'name':'ac30', 'is_dir': returnTrue}),
                type('obj', (object,), {'name':'ac27', 'is_dir': returnTrue}),
                type('obj', (object,), {'name':'ac26', 'is_dir': returnTrue})
            ], [
                type('obj', (object,), {'name':'b2300commands', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b2300types', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b2300utilities', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b2310commands', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b2310types', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b2310utilities', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b2305commands', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b2305types', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b2305utilities', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b1610commands', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b1610types', 'is_file': returnTrue}),
                type('obj', (object,), {'name':'b1610utilities', 'is_file': returnTrue}),
            ]
        ]
        
        mocked_os_scandir.side_effect = os_scandir_side_effect_values
        release, build = _Versioning.discover(53, 7000)
        self.assertEqual(release, 30)
        self.assertEqual(build, 2310)
        
        mocked_os_scandir.side_effect = os_scandir_side_effect_values
        release, build = _Versioning.discover(27, 2306)
        self.assertEqual(release, 27)
        self.assertEqual(build, 2305)
        
        mocked_os_scandir.side_effect = os_scandir_side_effect_values
        release, build = _Versioning.discover(28, 1610)
        self.assertEqual(release, 27)
        self.assertEqual(build, 1610)
        
        mocked_os_scandir.side_effect = os_scandir_side_effect_values
        self.assertRaises(StopIteration, _Versioning.discover, 23, 2300)
        
        mocked_os_scandir.side_effect = os_scandir_side_effect_values
        self.assertRaises(StopIteration, _Versioning.discover, 24, 1500)
        



"""Unittests for ACBaseType and inherited classes
"""

import unittest
from enum import Enum
from uuid import UUID
from typing import List, Optional
from archicad.acbasetype import _ACBaseType
from archicad import validators as V


VIEWMAP = 'ViewMap'
PROJECTMAP = 'ProjectMap'
LAYOUTBOOK = 'LayoutBook'
PUBLISHERSET = 'PublisherSets'



EXAMPLE_DICT = {'name': 'First Level',
                'type': LAYOUTBOOK,
                'example': {
                    'name': 'Second Level',
                    'type': LAYOUTBOOK,
                    'example': {
                        'name': 'Third Level',
                        'type': PROJECTMAP,
                        'example': {
                            'name': 'Ékezetek támogatása 😍',
                            'type': 'ProjectMap'
                        }
                    }
                }}

class TestACBaseType(unittest.TestCase):
    def test_base_instantiation(self):
        with self.assertRaises(TypeError):
            _ACBaseType({})

    def test_base_classinfo(self):
        self.assertIsNotNone(_ACBaseType.get_classinfo())
        self.assertEqual(len(_ACBaseType.get_classinfo().fields), 0)


class TestACTypes(unittest.TestCase):
    def setUp(self):
        class Example1(_ACBaseType):
            pass


        class Example2(_ACBaseType):
            __slots__ = ('type', 'name', 'example')

            def __init__(self, type, name, example=None):
                self.type: str = type
                self.name: str = name
                self.example: Optional[Example2] = example

        Example2.get_classinfo().add_field('type', str, V.value_set(['ViewMap', 'ProjectMap', 'LayoutBook', 'PublisherSets']))
        Example2.get_classinfo().add_field('name', str)
        Example2.get_classinfo().add_field('example', Optional[Example2])


        class Example3(_ACBaseType):
            __slots__ = ('name', 'guid', 'numbers', 'strings')

            def __init__(self, name, guid, numbers, strings):
                self.name: str = name
                self.guid: UUID = guid
                self.numbers: List[int] = numbers
                self.strings: List[Optional[List[str]]] = strings

        Example3.get_classinfo().add_field('name', str)
        Example3.get_classinfo().add_field('guid', UUID)
        Example3.get_classinfo().add_field('numbers', List[int], V.min_length(2), V.max_length(6))
        Example3.get_classinfo().add_field('strings', List[Optional[List[str]]])


        class Example4(_ACBaseType):
            __slots__ = ('items')

            def __init__(self, items):
                self.items: List[Example2] = items

        Example4.get_classinfo().add_field('items', List[Example2])
        self.Example1 = Example1
        self.Example2 = Example2
        self.Example3 = Example3
        self.Example4 = Example4
        self.ex1 = Example1()
        self.ex2 = Example2(type=PUBLISHERSET, name="The Publisher Set")
        self.ex3 = Example3('Sample Name', '260D42E3-AA2E-4DAD-BFFA-F2B5021501CC', [5, 6, 7], [])

    def test_inherited_instantiation(self):
        self.assertIsInstance(self.ex1, _ACBaseType)
        self.assertIsInstance(self.ex2, _ACBaseType)
        self.assertIsInstance(self.ex1, self.Example1)
        self.assertIsInstance(self.ex2, self.Example2)
        self.assertNotIsInstance(self.ex1, self.Example2)
        self.assertNotIsInstance(self.ex2, self.Example1)
        self.assertIs(type(self.ex1), self.Example1)
        self.assertIs(type(self.ex2), self.Example2)
        self.assertIsNot(type(self.ex1), _ACBaseType)
        self.assertIsNot(type(self.ex2), _ACBaseType)

    def test_classinfo(self):
        self.assertIsNotNone(self.Example1.get_classinfo())
        self.assertEqual(len(self.Example1.get_classinfo().fields), 0)
        self.assertIsNotNone(self.Example2.get_classinfo())
        self.assertEqual(len(self.Example2.get_classinfo().fields), 3)
        self.assertEqual(len(self.Example2.get_classinfo().value_validators), 1)
        self.assertIsNot(self.ex3.get_classinfo(), self.ex2.get_classinfo())

    def test_fields(self):
        self.assertIs(self.Example2.get_classinfo().fields['type'], str)
        self.assertIs(self.ex2.get_classinfo().fields['name'], str)
        self.assertIs(self.Example2.get_classinfo().fields['example'], Optional[self.Example2])
        self.assertIsNot(self.Example2.get_classinfo().fields['example'], _ACBaseType)

        with self.assertRaises(KeyError):
            self.assertIs(self.Example2.get_classinfo().fields['nothing'], str)
        with self.assertRaises(KeyError):
            self.ex1.new_attribute = 15
        with self.assertRaises(TypeError):
            self.Example2("Sample Name")

    def test_invalid_simple_attributes(self):
        self.assertIsNone(self.ex2.example)
        with self.assertRaises(TypeError):
            self.ex2.name = 5
        with self.assertRaises(TypeError):
            self.ex2.name = {'key': 'value'}
        with self.assertRaises(TypeError):
            self.ex2.type = 5
        with self.assertRaises(ValueError):
            self.ex2.type = 'Invalid'
        with self.assertRaises(ValueError):
            self.ex3.guid = '260D42E3-AA2E-4DAD-BFFA'
        with self.assertRaises(TypeError):
            self.ex3.guid = 25
        with self.assertRaises(TypeError):
            self.ex3.guid = None

    def test_valid_simple_attributes(self):
        self.ex2.name = 'Test Case'
        self.assertEqual(self.ex2.name, 'Test Case')
        self.ex2.type = 'ViewMap'
        self.assertEqual(self.ex2.type, VIEWMAP)
        self.ex3.guid = '260D42E3-AA2E-4DAD-BFFA-F2B5021501CC'
        self.assertEqual(self.ex3.guid, UUID('260D42E3-AA2E-4DAD-BFFA-F2B5021501CC'))


    def test_complex_attributes(self):
        with self.assertRaises(TypeError):
            self.ex2.example = 5
        with self.assertRaises(TypeError):
            self.ex2.example = {'example': True}
        with self.assertRaises(TypeError):
            self.ex2.example = self.Example1()
        with self.assertRaises(AttributeError):
            self.ex2.example.type

        self.ex2.example = {'name': 'Example Name', 'type': LAYOUTBOOK}
        self.assertIsNotNone(self.ex2.example)
        self.assertIsNone(self.ex2.example.example)
        self.assertEqual(self.ex2.example.name, 'Example Name')

        self.ex2.example = self.Example2(name='First', type=PROJECTMAP, example={'name': 'Second', 'type': LAYOUTBOOK})
        self.assertEqual(self.ex2.example.example.name, 'Second')
        self.assertEqual(self.ex2.example.type, PROJECTMAP)
        self.assertEqual(self.ex2.example.example.type, LAYOUTBOOK)

    def test_serialization(self):
        self.ex2.example = EXAMPLE_DICT
        self.assertEqual(self.ex2.example.example.example.name, 'Third Level')
        self.assertEqual(self.ex2.example.example.example.example.name, 'Ékezetek támogatása 😍')
        self.assertEqual(self.ex2.example.example.example.example.type, PROJECTMAP)

        result_dict = self.ex2.to_dict()
        self.assertEqual(len(result_dict), 3)
        self.assertDictEqual(result_dict['example'], EXAMPLE_DICT)
    
    def test_deserialization(self):
        myEx2 = self.Example2(**EXAMPLE_DICT)
        self.assertEqual(myEx2.name, 'First Level')
        self.assertEqual(myEx2.example.example.example.type, PROJECTMAP)
    
    def test_failed_deserialization(self):
        with self.assertRaises(TypeError):
            self.Example2(**{'name': 'Only Name'})
        with self.assertRaises(TypeError):
            self.Example2(**{'type': PROJECTMAP})
        with self.assertRaises(TypeError):
            self.Example2(**{'name': 15, 'type': PROJECTMAP})
        with self.assertRaises(TypeError):
            self.Example2(**{'name': 'Name', 'type': {'a': 1}})
        with self.assertRaises(TypeError):
            self.Example2(**{'name': 'Name', 'type': PROJECTMAP, 'invalid': 15})
        with self.assertRaises(TypeError):
            self.Example2(**{'name': 'Name', 'type': PROJECTMAP, 'example': {'name': 'Name2'}})
        with self.assertRaises(TypeError):
            self.Example2(**{'name': 'Name', 'type': PROJECTMAP, 'example': {'name': 'Name', 'type': 'Invalid'}})
        with self.assertRaises(ValueError):
            self.Example2(**{'name': 'Name', 'type': 'Invalid'})

    def test_nested_lists(self):
        self.ex3 = self.Example3(guid='260D42E3-AA2E-4DAD-BFFA-F2B5021501CC',
                            numbers=[2, 4, 6, 8.4],
                            strings=[['a', 'b', 'c'], ['d', 'e', 'f'], ['g', 'h', 'i']],
                            name='Name')
        self.assertEqual(len(self.ex3.numbers), 4)
        self.assertEqual(len(self.ex3.strings), 3)
        self.assertListEqual(self.ex3.strings[1], ['d', 'e', 'f'])
        self.assertDictEqual(self.ex3.to_dict(), {'name': 'Name',
                                                  'guid': '260D42E3-AA2E-4DAD-BFFA-F2B5021501CC',
                                                  'numbers': [2, 4, 6, 8],
                                                  'strings': [['a', 'b', 'c'], ['d', 'e', 'f'], ['g', 'h', 'i']]})

    def test_list_items(self):
        ex4 = self.Example4([self.ex2])
        self.assertEqual(len(ex4.items), 1)

    def test_number_initialization(self):
        class ExampleNumbers(_ACBaseType):
            __slots__ = ('guid', 'width', 'height', 'amount')

            def __init__(self, guid, width, height, amount):
                self.guid: UUID = guid
                self.width: float = width
                self.height: float = height
                self.amount: int = amount

        ExampleNumbers.get_classinfo().add_field('guid', UUID)
        ExampleNumbers.get_classinfo().add_field('width', float)
        ExampleNumbers.get_classinfo().add_field('height', float)
        ExampleNumbers.get_classinfo().add_field('amount', int, V.minimum(1, False))

        e1 = ExampleNumbers('260D42E3-AA2E-4DAD-BFFA-F2B5021501CC', 1.0, 2.5, 3)
        self.assertIsInstance(e1.width, float)
        self.assertIsInstance(e1.height, float)
        self.assertIsInstance(e1.amount, int)
        
        e2 = ExampleNumbers('260D42E3-AA2E-4DAD-BFFA-F2B5021501CC', 1, 2.5, 3)
        self.assertIsInstance(e2.width, float)
        self.assertIsInstance(e2.height, float)
        self.assertIsInstance(e2.amount, int)
        
        with self.assertRaises(TypeError):
            e3 = ExampleNumbers('260D42E3-AA2E-4DAD-BFFA-F2B5021501CC', 1.0, 2.5, 3.5)


import unittest
from unittest.mock import patch, ANY

from archicad.commands import UnsucceededCommandCall
from archicad.connection import ACConnection


class TestACConnection(unittest.TestCase):
    def setUp(self):
        self.prod_info_patcher = patch('archicad.connection.BasicCommands.GetProductInfo')
        self.versioning_patcher = patch('archicad.connection._Versioning')
        self.mocked_prod_info = self.prod_info_patcher.start()
        self.mocked_versioning = self.versioning_patcher.start()
        self.mocked_prod_info.return_value = [24, 2310, 'INT']
        self.mocked_versioning.return_value = type('obj', (object,), {'commands':1, 'types':2, 'utilities':3})
        
    def tearDown(self):
        self.prod_info_patcher.stop()
        self.versioning_patcher.stop()

    def test_connect_on_port(self):
        self.assertRaises(AssertionError, ACConnection.connect, 1500)
        self.assertEqual(self.mocked_prod_info.called, False)
        self.assertEqual(self.mocked_versioning.called, False)

        self.assertRaises(AssertionError, ACConnection.connect, 19722)
        self.assertEqual(self.mocked_prod_info.called, False)
        self.assertEqual(self.mocked_versioning.called, False)

        self.assertIsInstance(ACConnection.connect(19723), ACConnection)
        self.mocked_prod_info.assert_called_with()
        self.mocked_versioning.assert_called_with(24, 2310, ANY)

        self.assertIsInstance(ACConnection.connect(19728), ACConnection)
        self.mocked_prod_info.assert_called_with()
        self.mocked_versioning.assert_called_with(24, 2310, ANY)

        self.assertIsInstance(ACConnection.connect(19743), ACConnection)
        self.mocked_prod_info.assert_called_with()
        self.mocked_versioning.assert_called_with(24, 2310, ANY)

        self.assertRaises(AssertionError, ACConnection.connect, 19744)

    @patch('archicad.connection.ACConnection.port_from_args')
    def test_connect_with_port_from_args(self, mocked_port_from_args):
        mocked_port_from_args.return_value = 2140
        self.assertRaises(AssertionError, ACConnection.connect)
        self.assertEqual(self.mocked_prod_info.called, False)
        self.assertEqual(self.mocked_versioning.called, False)

        mocked_port_from_args.return_value = 19722
        self.assertRaises(AssertionError, ACConnection.connect)
        self.assertEqual(self.mocked_prod_info.called, False)
        self.assertEqual(self.mocked_versioning.called, False)

        mocked_port_from_args.return_value = 19723
        self.assertIsInstance(ACConnection.connect(), ACConnection)
        self.mocked_prod_info.assert_called_with()
        self.mocked_versioning.assert_called_with(24, 2310, ANY)

        mocked_port_from_args.return_value = 19728
        self.assertIsInstance(ACConnection.connect(), ACConnection)
        self.mocked_prod_info.assert_called_with()
        self.mocked_versioning.assert_called_with(24, 2310, ANY)

        mocked_port_from_args.return_value = 19742
        self.assertIsInstance(ACConnection.connect(), ACConnection)
        self.mocked_prod_info.assert_called_with()
        self.mocked_versioning.assert_called_with(24, 2310, ANY)

        
    @patch('archicad.connection.ACConnection.port_from_args')
    def test_connect_on_default_port(self, mocked_port_from_args):
        self.assertIs(mocked_port_from_args, ACConnection.port_from_args)
        mocked_port_from_args.return_value = None

        conn = ACConnection.connect()
        self.mocked_prod_info.assert_called_with()
        self.mocked_versioning.assert_called_with(24, 2310, ANY)
        self.assertIsInstance(conn, ACConnection)
        self.assertEqual(conn.port, 19723)

        self.mocked_prod_info.side_effect = [
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            [24, 2310, 'INT'],
            [24, 2310, 'INT']
        ]
        conn = ACConnection.connect()
        self.mocked_prod_info.assert_called_with()
        self.mocked_versioning.assert_called_with(24, 2310, ANY)
        self.assertIsInstance(conn, ACConnection)
        self.assertEqual(conn.port, 19727)


        self.mocked_prod_info.side_effect = [
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            [24, 2310, 'INT'],
            [24, 2310, 'INT']
        ]
        conn = ACConnection.connect()
        self.mocked_prod_info.assert_called_with()
        self.mocked_versioning.assert_called_with(24, 2310, ANY)
        self.assertIsInstance(conn, ACConnection)
        self.assertEqual(conn.port, 19743)
        
        self.mocked_prod_info.side_effect = [
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall(),
            UnsucceededCommandCall()
        ]
        self.assertIsNone(ACConnection.connect())
        self.mocked_prod_info.assert_called_with()
        self.mocked_versioning.assert_called_with(24, 2310, ANY)

